#include "plotdata.h"
#include "TMath.h"

#include "TStyle.h"
#include "TAxis.h"
#include "TFile.h"
#include <fstream>
#include <iostream>
#include "omp.h"
#include "QDebug"
#include "QDateTime"

PlotData::PlotData(const Setup* config_)
      : config(config_) 
{ 
    counts = 0;
    probDist = std::vector<std::vector<double>>(config->thetaBins, std::vector<double>(config->phiBins, 0));
    pixelateSphere();

    hist = new TH2D("ROI", " ; Azimuth (degree); Elevation (degree)",
                        config->phiBins, config->phiMin * 180 / M_PI, config->phiMax * 180 / M_PI,
                        config->thetaBins, config->thetaMin * 180 / M_PI, config->thetaMax * 180 / M_PI);
    hist->Reset();
    gStyle->SetOptStat(0);
    hist->GetZaxis()->SetLabelSize(0.02);
    
    spectrum = new TH1D("Spectrum", " ; Energy (MeV); Counts",
                        config->ergBins, config->ergMin, config->ergMax);
    spectrum->Reset();
    gStyle->SetOptStat(0);
    spectrum->SetFillColor(40);

    std::string strtmp = "Total counts: " + std::to_string(counts);
    countsText = new TText(0.7, 0.92, strtmp.c_str());
}

PlotData::~PlotData()
{
    if (hist){
        delete hist;
        hist = nullptr;
    }
    if (spectrum){
        delete spectrum;
        spectrum = nullptr;
    }
    if (countsText){
        delete countsText;
        countsText = nullptr;
    }
    for (std::size_t i = 0; i < longitudes.size(); ++i) {
       if(longitudes[i]) {
           delete longitudes[i];
           longitudes[i] = nullptr;
       }
    }
    for (std::size_t i = 0; i < latitudes.size(); ++i) {
       if (latitudes[i]) {
           delete latitudes[i];
           latitudes[i] = nullptr;
       }
    }
}

void PlotData::unZoomImage()
{
    mMutex.lock();
    ((TH1*)(hist))->GetXaxis()->UnZoom();
    ((TH1*)(hist))->GetYaxis()->UnZoom();
    mMutex.unlock();
}

void PlotData::unZoomSpectrum()
{
    mMutex.lock();
    spectrum->GetXaxis()->UnZoom();
    spectrum->GetYaxis()->UnZoom();
    mMutex.unlock();
}
void PlotData::rebinImage()
{
    mMutex.lock();
    counts = 0;
    hist->SetBins(config->phiBins, config->phiMin * 180 / M_PI, config->phiMax * 180 / M_PI,
                  config->thetaBins, config->thetaMin * 180 / M_PI, config->thetaMax * 180 / M_PI);
    probDist = std::vector<std::vector<double>>(config->thetaBins, std::vector<double>(config->phiBins, 0));
    pixelateSphere();
    mMutex.unlock();
}

void PlotData::rebinSpectrum()
{
    mMutex.lock();
    hist->SetBins(config->ergBins, config->ergMin, config->ergMax);
    mMutex.unlock();
//    return;
}

int PlotData::pixelateSphere() {
    xbs = std::vector<std::vector<Vector3D>>(config->thetaBins, std::vector<Vector3D>(config->phiBins, Vector3D(0, 0, 0)));
    thetaBinCenters = std::vector<double>(config->thetaBins, 0);
    phiBinCenters = std::vector<double>(config->phiBins, 0);
    dphi = (config->phiMax - config->phiMin) / config->phiBins;
    dtheta = (config->thetaMax - config->thetaMin) / config->thetaBins;
    double phi = config->phiMin + dphi / 2;;
    double theta = config->thetaMin + dtheta / 2;
    for (int i = 0; i < config->thetaBins; i++)
    {
        thetaBinCenters[i] = theta;
        phi = config->phiMin + dphi / 2;
        for (int j = 0; j < config->phiBins; j++)
        {
            if (i==0)
                phiBinCenters[j] = phi;
            xbs[i][j].X = config->R * std::cos(theta) * std::cos(phi);
            xbs[i][j].Y = config->R * std::cos(theta) * std::sin(phi);
            xbs[i][j].Z = config->R * std::sin(theta);
            phi += dphi;
        }
        theta += dtheta;
    }
    return 0;
}

void PlotData::drawImage()
{
//    hist->Draw("z aitoff");
    hist->Draw("COLZ");
}

void PlotData::drawSpectrum()
{
    spectrum->Draw();
}

void PlotData::saveOnExit()
{
    QDateTime dateTime = QDateTime::currentDateTime();
    QString currentDateTime = dateTime.toString("yyyy-MM-dd-HH-mm-ss");
    mMutex.lock();
    if (config->saveImage)
    {
        if (config->saveImageFormat == "ROOT"){
            QString filename = config->outputDir.absolutePath() + QString("/Image-%1.root").arg(currentDateTime);
            TFile f(filename.toLocal8Bit().constData(),"new");
            hist->Write();
            f.Close();
        }
        else if (config->saveImageFormat == "TXT"){
            QString filename = config->outputDir.absolutePath() + QString("/Image-%1.txt").arg(currentDateTime);
            hist2txt(filename.toStdString());
        }
    }
    if (config->saveSpectrum)
    {
        if (config->saveSpectrumFormat == "ROOT"){
            QString filename = config->outputDir.absolutePath() + QString("/Spectrum-%1.root").arg(currentDateTime);
//            spectrum->SaveAs(filename.toLocal8Bit().constData(), "root");
            TFile f(filename.toLocal8Bit().constData(),"new");
            spectrum->Write();
            f.Close();
        }
        else if (config->saveImageFormat == "TXT"){
            QString filename = config->outputDir.absolutePath() + QString("/Spectrum-%1.txt").arg(currentDateTime);
            spectrum2txt(filename.toStdString());
        }
    }
    mMutex.unlock();
}

void PlotData::clearAll()
{
    mMutex.lock();
    counts=0;
    hist->Reset();
    spectrum->Reset();
    mMutex.unlock();
}

void PlotData::clearImage()
{
    mMutex.lock();
    counts=0;
    hist->Reset();
    mMutex.unlock();
}

void PlotData::clearSpectrum()
{
    mMutex.lock();
    spectrum->Reset();
    mMutex.unlock();
}

bool PlotData::saveImage2txt(const std::string& fileName)
{
//        qDebug() << "Save as text file";
    mMutex.lock();
    bool saved=hist2txt(fileName);
    mMutex.unlock();
    return saved;
}

bool PlotData::hist2txt(const std::string& fileName)
{
    std::ofstream outf(fileName.c_str(), std::ios::out);
    if (!outf.good())
    {
        return false;
    }

    outf << "     Phi   Theta      Content\n";
    for (int i = 1; i <= hist->GetNbinsX(); i++)
    {
        for (int j = 1; j <= hist->GetNbinsY(); j++)
        {
            outf << std::fixed    << std::setprecision(2)
                 << std::setw(8)  << ((TAxis*)hist->GetXaxis())->GetBinCenter(i)
                 << std::setw(8)  << ((TAxis*)hist->GetYaxis())->GetBinCenter(j)
                 << std::fixed    << std::setprecision(8)
                 << std::setw(13) << hist->GetBinContent(i,j) << '\n';
        }
    }
    outf.close();
//    qDebug() << "Text file saved.";
    return true;
}

bool PlotData::saveSpectrum2txt(const std::string& fileName)
{
//        qDebug() << "Save as text file";
    mMutex.lock();
    bool saved=spectrum2txt(fileName);
    mMutex.unlock();
    return saved;
}

bool PlotData::spectrum2txt(const std::string& fileName)
{
    std::ofstream outf(fileName.c_str(), std::ios::out);
    if (!outf.good())
    {
        return false;
    }

    outf << "     BinCenter      Counts\n";
    for (int i = 1; i <= spectrum->GetNbinsX(); i++)
    {
        outf << std::fixed    << std::setprecision(2)
             << std::setw(8)  << spectrum->GetBinCenter(i)
             << std::setw(8)  << spectrum->GetBinContent(i)<< '\n';
    }
    outf.close();
//    qDebug() << "Text file saved.";
    return true;
}

void PlotData::updateImage(std::vector<Cone>::const_iterator first,
                                 std::vector<Cone>::const_iterator last,
                                 const bool normalized)
{
    if (first==last)
        return;
    else if (normalized) {
        mMutex.lock();
        addConesNormalized(first, last);
        mMutex.unlock();
    }
    else {
        mMutex.lock();
        addCones(first, last);
        mMutex.unlock();
    }
}

int PlotData::addCones(std::vector<Cone>::const_iterator first,
             std::vector<Cone>::const_iterator last)
{
    // project cones onto the spherical surface
    double alpha(0);
    double beta(0);
    double E2(0);
    double sgma2(0);
    double sgmb2(0);
    Vector3D ray;
    for (auto k = first; k < last; k++)
    {
        alpha = k->cosHalfAngle;
        spectrum->Fill(k->E0);
        // ideal case: inital energy is known
        // sgma2 = std::pow(0.511*k->E0/std::pow((k->E0-k->Edpst),2)*config->sgmE, 2);
        // realistic case: initial energy = E_1 + E_2
        E2 = k->E0 - k->Edpst;
        sgma2 = std::pow(k->Edpst/std::pow(k->E0, 2) , 2) + std::pow(1/E2 - E2/std::pow(k->E0, 2) , 2);
        sgma2 *= std::pow(config->sgmE * 0.511, 2);
        #pragma omp parallel for private(ray, beta, sgmb2) collapse(2)
        for (int i = 0; i < config->thetaBins; i++)
        {
            for (int j = 0; j < config->phiBins; j++)
            {
                ray = k->apex - xbs[i][j];
                beta = getCosAngle(ray, k->axis);
                sgmb2 = std::pow((ray/(ray*ray) + k->axis/(k->axis*k->axis)-
                                (ray+k->axis)/(ray*k->axis))*config->sgmpos * beta, 2);
                sgmb2+= std::pow((ray/(ray*k->axis)-k->axis/(k->axis*k->axis))*config->sgmpos * beta, 2);
                hist->SetBinContent(j+1, i+1,
                    hist->GetBinContent(j+1, i+1) +
                    std::exp(-std::pow((std::pow(beta, config->order)-std::pow(alpha, config->order)), 2)/
                             (2*std::pow(config->order, 2)*
                              (std::pow(alpha,2*config->order-2)*sgma2 +std::pow(beta, 2*config->order-2)*sgmb2))));
            }
        }
    }
    counts += (last - first);
    return 0;
}

//int PlotData::addConesNormalized(std::vector<Cone>::const_iterator first,
//             std::vector<Cone>::const_iterator last)
//{
//    // project cones onto the spherical surface
//    double alpha(0);
//    double beta(0);
//    double E2(0);
//    double sgma2(0);
//    double sgmb2(0);
//    double summ(0);
//    Vector3D ray;
//    for (auto k = first; k < last; k++)
//    {
//        counts++;
//        if (counts == 3507)
//            qDebug() << "here";
//        alpha = k->cosHalfAngle;
//        spectrum->Fill(k->E0);
//        // ideal case: inital energy is known
//        // sgma2 = std::pow(0.511*k->E0/std::pow((k->E0-k->Edpst),2)*config->sgmE, 2);
//        // realistic case: initial energy = E_1 + E_2
//        E2 = k->E0 - k->Edpst;
//        sgma2 = std::pow(k->Edpst/std::pow(k->E0, 2) , 2) + std::pow(1/E2 - E2/std::pow(k->E0, 2) , 2);
//        sgma2 *= std::pow(config->sgmE * 0.511, 2);
//        summ = 0;
//        #pragma omp parallel for private(ray, beta, sgmb2) shared(probDist) reduction(+:summ)
//        for (int i = 0; i < config->thetaBins; i++)
//        {
//            for (int j = 0; j < config->phiBins; j++)
//            {
//                ray = k->apex - xbs[i][j];
//                beta = getCosAngle(ray, k->axis);
//                sgmb2 = std::pow((ray/(ray*ray) + k->axis/(k->axis*k->axis)-
//                                (ray+k->axis)/(ray*k->axis))*config->sgmpos * beta, 2);
//                sgmb2+= std::pow((ray/(ray*k->axis)-k->axis/(k->axis*k->axis))*config->sgmpos * beta, 2);
//                probDist[i][j] = std::exp(-std::pow((std::pow(beta, config->order)-std::pow(alpha, config->order)), 2)/
//                                (2*std::pow(config->order, 2)*
//                                (std::pow(alpha,2*config->order-2)*sgma2 +std::pow(beta, 2*config->order-2)*sgmb2)));
//                // use proability density
//                summ += probDist[i][j] * dtheta * dphi * std::cos(thetaBinCenters[i]);
//                // uncomment to use integral in the bin
//                // probDist[i][j] = probDist[i][j] * config->dtheta * config->dphi * std::cos(config->thetaBinCenters[i]);
//                // summ += probDist[i][j];
//            }
//        }
//        // #pragma omp parallel for
//        for (int i = 0; i < config->thetaBins; i++)
//        {
//            for (int j = 0; j < config->phiBins; j++)
//            {
//                hist->SetBinContent(j+1, i+1, (hist->GetBinContent(j+1, i+1)*(counts - 1) + probDist[i][j] / summ) / counts);
//                if(std::isnan(hist->GetBinContent(j+1, i+1)) || std::isinf(hist->GetBinContent(j+1,i+1)))
//                    qDebug() << i << j << counts;
//            }
//        }
//    }
//    return 0;
//}

int PlotData::addConesNormalized(std::vector<Cone>::const_iterator first,
             std::vector<Cone>::const_iterator last)
{
    // project cones onto the spherical surface
    double alpha(0);
    double beta(0);
    double E2(0);
    double sgma2(0);
//    double sgmb2(0);
//    double summ(0);
    Vector3D ray;
    double A(0);
    for (auto k = first; k < last; k++)
    {
        counts++;
        alpha = k->cosHalfAngle;
        spectrum->Fill(k->E0);
        // ideal case: inital energy is known
        // sgma2 = std::pow(0.511*k->E0/std::pow((k->E0-k->Edpst),2)*config->sgmE, 2);
        // realistic case: initial energy = E_1 + E_2
        E2 = k->E0 - k->Edpst;
        sgma2 = std::pow(k->Edpst/std::pow(k->E0, 2) , 2) + std::pow(1/E2 - E2/std::pow(k->E0, 2) , 2);
        sgma2 *= std::pow(config->sgmE * 0.511, 2);
        A = 0.5 * M_1_PI * M_2_SQRTPI * M_SQRT1_2 / std::sqrt(sgma2) / (std::erf((1-alpha)/ (M_SQRT2 * std::sqrt(sgma2))) - std::erf((-1-alpha)/ (M_SQRT2 * std::sqrt(sgma2))));
//        #pragma omp parallel for private(ray, beta, sgmb2) shared(probDist) reduction(+:summ)
        for (int i = 0; i < config->thetaBins; i++)
        {
            for (int j = 0; j < config->phiBins; j++)
            {
                ray = k->apex - xbs[i][j];
                beta = getCosAngle(ray, k->axis);

                probDist[i][j] = A * std::exp(-std::pow(beta-alpha, 2)/
                                (2*sgma2));
            }
        }
        // #pragma omp parallel for
        for (int i = 0; i < config->thetaBins; i++)
        {
            for (int j = 0; j < config->phiBins; j++)
            {
                hist->SetBinContent(j+1, i+1, (hist->GetBinContent(j+1, i+1)*(counts-1) + probDist[i][j]) / counts);
//                hist->SetBinContent(j+1, i+1, (hist->GetBinContent(j+1, i+1)*(counts - 1) + probDist[i][j] / summ) / counts);
                if(std::isnan(hist->GetBinContent(j+1, i+1)) || std::isinf(hist->GetBinContent(j+1,i+1)))
                    qDebug() << i << j << counts;
            }
        }
    }
    return 0;
}


//void PlotData::initGridlines()
//{
//    const int Nl = 19; // Number of drawn latitudes
//    const int NL = 19; // Number of drawn longitudes
//    const int M = 30;
//    float conv=M_PI/180;
//    float la, lo, x, y, z;
//    if (latitudes.size()==0)
//    {
//        for (int j=0;j<Nl;++j) {
//            // latitudes[j]=new TGraph();
//            latitudes.push_back(new TGraph());
//            la = -90+180/(Nl-1)*j;
//            for (int i=0;i<M+1;++i) {
//                lo = -180+360/M*i;
//                z  = sqrt(1+cos(la*conv)*cos(lo*conv/2));
//                x  = 180*cos(la*conv)*sin(lo*conv/2)/z;
//                y  = 90*sin(la*conv)/z;
//                latitudes[j]->SetPoint(i,x,y);
//            }
//        }
//    }

//    if (longitudes.size()==0)
//    {
//        for (int j=0;j<NL;++j) {
//            // longitudes[j]=new TGraph();
//            longitudes.push_back(new TGraph());
//            lo = -180+360/(NL-1)*j;
//            for (int i=0;i<M+1;++i) {
//                la = -90+180/M*i;
//                z  = sqrt(1+cos(la*conv)*cos(lo*conv/2));
//                x  = 180*cos(la*conv)*sin(lo*conv/2)/z;
//                y  = 90*sin(la*conv)/z;
//                longitudes[j]->SetPoint(i,x,y);
//            }
//        }
//    }

//    // Draw the grid
//    TPad* pad2 = new TPad("pad2","",0,0,1,1);
//    pad2->SetFillStyle(4000);
//    pad2->SetFillColor(0);
//    pad2->SetBorderSize(0);
//    pad2->SetFrameBorderMode(0);
//    pad2->SetFrameLineColor(0);
//    pad2->SetFrameBorderMode(0);
//    pad2->Draw();
//    pad2->cd();
//    Double_t ymin = config->thetaMin * 180 / M_PI; //-89.5;
//    Double_t ymax = config->thetaMax * 180 / M_PI; //89.5;
//    Double_t dy = (ymax-ymin)/0.8; //10 per cent margins top and bottom
//    Double_t xmin = config->phiMin * 180 / M_PI; //-180;
//    Double_t xmax = config->phiMax * 180 / M_PI; //180;
//    Double_t dx = (xmax-xmin)/0.8; //10 per cent margins left and right

//    pad2->Range(xmin-0.1*dx,ymin-0.1*dy,xmax+0.1*dx,ymax+0.1*dy);

//    for (int j=0;j<Nl;++j) latitudes[j]->Draw("l");
//    for (int j=0;j<NL;++j) longitudes[j]->Draw("l");
//}

void PlotData::drawCountsLabel()
{
    std::string strtmp = "Total counts: " + std::to_string(counts);
    countsText->SetText(0.7, 0.92, strtmp.c_str());
    countsText->SetTextSizePixels(5);
    countsText->SetNDC(kTRUE);
    countsText->Draw("SAME");
}

void PlotData::aitoff2xy(const double& l, const double& b, double &Al, double &Ab)
{
    Double_t x, y;
    Double_t alpha2 = (l/2)*TMath::DegToRad();
    Double_t delta  = b*TMath::DegToRad();
    Double_t r2     = TMath::Sqrt(2.);
    Double_t f      = 2*r2/TMath::Pi();
    Double_t cdec   = TMath::Cos(delta);
    Double_t denom  = TMath::Sqrt(1. + cdec*TMath::Cos(alpha2));
    x      = cdec*TMath::Sin(alpha2)*2.*r2/denom;
    y      = TMath::Sin(delta)*r2/denom;
    x     *= TMath::RadToDeg()/f;
    y     *= TMath::RadToDeg()/f;
    //  x *= -1.; // for a skymap swap left<->right
    Al = x;
    Ab = y;
}
