#include "plotwindow.h"
#include <QDebug>
#include <QDateTime>

QPlotWindow::QPlotWindow(QWidget *parent, const Setup* config_, PlotData* plotdata_)
    : QMainWindow(parent),
      config(config_),
      ui(new Ui::PlotWindow),
      plotdata(plotdata_)
{
    ui->setupUi(this);
    ui->imageCanvas->setID(1);
    ui->spectrumCanvas->setID(2);

    connect(ui->actionStart, &QAction::triggered, this, &QPlotWindow::handleStart);
    connect(ui->actionPause, &QAction::triggered, this, &QPlotWindow::handlePause);
    connect(ui->actionStop, &QAction::triggered, this, &QPlotWindow::handleStop);
    connect(ui->actionUnzoom, &QAction::triggered, this, &QPlotWindow::handleUnzoom);
    connect(ui->actionClearAll, &QAction::triggered, this, &QPlotWindow::handleClear);

    connect(ui->imageCanvas, &QCustomCanvas::canvasZoomedout, this, &QPlotWindow::handleUnzoom);
    connect(ui->spectrumCanvas, &QCustomCanvas::canvasZoomedout, this, &QPlotWindow::handleUnzoom);
    connect(ui->imageCanvas, &QCustomCanvas::canvasCleared, this, &QPlotWindow::handleClear);
    connect(ui->spectrumCanvas, &QCustomCanvas::canvasCleared, this, &QPlotWindow::handleClear);
    connect(ui->imageCanvas, &QCustomCanvas::canvasSaved2Image, this, &QPlotWindow::handleScreenshot);
    connect(ui->spectrumCanvas, &QCustomCanvas::canvasSaved2Image, this, &QPlotWindow::handleScreenshot);
    connect(ui->imageCanvas, &QCustomCanvas::canvasSaved2Txt, this, &QPlotWindow::handleSave2txt);
    connect(ui->spectrumCanvas, &QCustomCanvas::canvasSaved2Txt, this, &QPlotWindow::handleSave2txt);

    // draw image
    ui->imageCanvas->Canvas()->cd();
    plotdata->mMutex.lock();
    plotdata->drawImage();
//    plotdata->initGridlines();
    plotdata->drawCountsLabel();
    plotdata->mMutex.unlock();
    ui->imageCanvas->Canvas()->Modified();
    ui->imageCanvas->Canvas()->Update();

    // draw spectrum
    ui->spectrumCanvas->Canvas()->cd();
    plotdata->mMutex.lock();
    plotdata->drawSpectrum();
    plotdata->mMutex.unlock();
    ui->spectrumCanvas->Canvas()->Modified();
    ui->spectrumCanvas->Canvas()->Update();
}


QPlotWindow::~QPlotWindow()
{
    delete ui;
}

void QPlotWindow::handleStart(){
    emit runStarted();
    ui->statusbar->showMessage("Running.");
}
void QPlotWindow::handlePause()
{
    emit runPaused();
    ui->statusbar->showMessage("Paused.");
}
void QPlotWindow::handleStop()
{
    emit runStopped();
    ui->statusbar->showMessage("Stopped.");
}

void QPlotWindow::handleUnzoom(int id)
{
    if (id==0)
    {
        plotdata->unZoomImage();
        plotdata->unZoomSpectrum();
    }
    else if (id ==1)
    {
        plotdata->unZoomImage();
    }
    else if (id ==2)
    {
        plotdata->unZoomSpectrum();
    }
}

void QPlotWindow::reopen()
{
    ui->imageCanvas->Canvas()->cd();
    ui->imageCanvas->Canvas()->Modified();
    ui->imageCanvas->Canvas()->Update();

    ui->spectrumCanvas->Canvas()->cd();
    ui->spectrumCanvas->Canvas()->Modified();
    ui->spectrumCanvas->Canvas()->Update();
}

void QPlotWindow::changeEvent(QEvent *e)
{
    if(e->type() == QEvent::WindowStateChange){
        QWindowStateChangeEvent *event = static_cast<QWindowStateChangeEvent *>(e);
        if((event->oldState() & Qt::WindowMaximized)||
                (event->oldState() & Qt::WindowMinimized)||
                (event->oldState() & Qt::WindowNoState &&
                 this->windowState() == Qt::WindowMaximized)){
            if(ui->imageCanvas->Canvas()){
                ui->imageCanvas->Canvas()->Resize();
                ui->imageCanvas->Canvas()->Update();
            }
            if(ui->spectrumCanvas->Canvas()){
                ui->spectrumCanvas->Canvas()->Resize();
                ui->spectrumCanvas->Canvas()->Update();
            }
        }
    }
}

void QPlotWindow::redraw()
{
    plotdata->mMutex.lock();
    ui->imageCanvas->Canvas()->cd();
    std::string strtmp = "Total counts: " + std::to_string(plotdata->counts);
    plotdata->getCountsText()->SetText(0.7, 0.92, strtmp.c_str());
    ui->imageCanvas->Canvas()->Modified();
    ui->imageCanvas->Canvas()->Update();

    ui->spectrumCanvas->Canvas()->cd();
    ui->spectrumCanvas->Canvas()->Modified();
    ui->spectrumCanvas->Canvas()->Update();
    plotdata->mMutex.unlock();
}

void QPlotWindow::handleClear(int id){
    if (id==0){
        qDebug() << "Clear all";
        plotdata->clearAll();
    }
    else if (id ==1) {
        qDebug() << "Clear image";
        plotdata->clearImage();
    }
    else if (id ==2) {
        qDebug() << "Clear spectrum";
        plotdata->clearSpectrum();
    }
}

void QPlotWindow::handleScreenshot(int id){
    QDateTime dateTime = QDateTime::currentDateTime();
    QString currentDateTime = dateTime.toString("yyyy-MM-dd-HH-mm-ss");
    if (id==0){
        qDebug() << "Save all";
        handleScreenshot(1);
        handleScreenshot(2);
    }
    else if (id ==1) {
        qDebug() << "Save image";
        QString fileName = config->outputDir.absolutePath() + QString("/Screenshot-Image-%1.png").arg(currentDateTime);
        plotdata->mMutex.lock();
        ui->imageCanvas->Canvas()->SaveAs(fileName.toLocal8Bit().constData());
        plotdata->mMutex.unlock();
        ui->statusbar->showMessage(QString("Image saved to: %1").arg(fileName));
    }
    else if (id ==2) {
        qDebug() << "Save spectrum";
        QString fileName = config->outputDir.absolutePath() + QString("/Screenshot-Spctrum-%1.png").arg(currentDateTime);
        plotdata->mMutex.lock();
        ui->spectrumCanvas->Canvas()->SaveAs(fileName.toLocal8Bit().constData());
        plotdata->mMutex.unlock();
        ui->statusbar->showMessage(QString("Spectrum saved to: %1").arg(fileName));
    }
}

void QPlotWindow::handleSave2txt(int id){
    QDateTime dateTime = QDateTime::currentDateTime();
    QString currentDateTime = dateTime.toString("yyyy-MM-dd-HH-mm-ss");
    if (id==0){
        qDebug() << "Save all";
    }
    else if (id ==1) {
        qDebug() << "Save image to txt";
        QString fileName = config->outputDir.absolutePath() + QString("/Image-%1.txt").arg(currentDateTime);
        plotdata->saveImage2txt(fileName.toStdString());
        ui->statusbar->showMessage(QString("Image saved to: %1").arg(fileName));
    }
    else if (id ==2) {
        qDebug() << "Save spectrum to txt";
        QString fileName = config->outputDir.absolutePath() + QString("/Spectrum-%1.txt").arg(currentDateTime);
        plotdata->saveSpectrum2txt(fileName.toStdString());
        ui->statusbar->showMessage(QString("Spectrum saved to: %1").arg(fileName));
    }
}
